@@ -22,6 +22,7 @@ import json
 import subprocess
 from ..mesonlib import MesonException
 from ..mesonlib import get_compiler_for_source, classify_unity_sources
+from ..mesonlib import File
 from ..compilers import CompilerArgs
 from collections import OrderedDict
 import shlex
@@ -89,12 +90,17 @@ class OptionProxy:
 class OptionOverrideProxy:
     '''Mimic an option list but transparently override
     selected option values.'''
-    def __init__(self, overrides, options):
+    def __init__(self, overrides, *options):
         self.overrides = overrides
         self.options = options
 
     def __getitem__(self, option_name):
-        base_opt = self.options[option_name]
+        for opts in self.options:
+            if option_name in opts:
+                return self._get_override(option_name, opts[option_name])
+        raise KeyError('Option not found', option_name)
+
+    def _get_override(self, option_name, base_opt):
         if option_name in self.overrides:
             return OptionProxy(base_opt.name, base_opt.validate_value(self.overrides[option_name]))
         return base_opt
@@ -108,9 +114,6 @@ class Backend:
         self.processed_targets = {}
         self.build_to_src = os.path.relpath(self.environment.get_source_dir(),
                                             self.environment.get_build_dir())
-        for t in self.build.targets:
-            priv_dirname = self.get_target_private_dir_abs(t)
-            os.makedirs(priv_dirname, exist_ok=True)
 
     def get_target_filename(self, t):
         if isinstance(t, build.CustomTarget):
@@ -126,6 +129,20 @@ class Backend:
     def get_target_filename_abs(self, target):
         return os.path.join(self.environment.get_build_dir(), self.get_target_filename(target))
 
+    def get_builtin_options_for_target(self, target):
+        return OptionOverrideProxy(target.option_overrides,
+                                   self.environment.coredata.builtins)
+
+    def get_base_options_for_target(self, target):
+        return OptionOverrideProxy(target.option_overrides,
+                                   self.environment.coredata.builtins,
+                                   self.environment.coredata.base_options)
+
+    def get_compiler_options_for_target(self, target):
+        return OptionOverrideProxy(target.option_overrides,
+                                   # no code depends on builtins for now
+                                   self.environment.coredata.compiler_options)
+
     def get_option_for_target(self, option_name, target):
         if option_name in target.option_overrides:
             override = target.option_overrides[option_name]
@@ -170,12 +187,10 @@ class Backend:
         return self.build_to_src
 
     def get_target_private_dir(self, target):
-        dirname = os.path.join(self.get_target_dir(target), target.get_basename() + target.type_suffix())
-        return dirname
+        return os.path.join(self.get_target_dir(target), target.get_id())
 
     def get_target_private_dir_abs(self, target):
-        dirname = os.path.join(self.environment.get_build_dir(), self.get_target_private_dir(target))
-        return dirname
+        return os.path.join(self.environment.get_build_dir(), self.get_target_private_dir(target))
 
     def get_target_generated_dir(self, target, gensrc, src):
         """
@@ -191,6 +206,8 @@ class Backend:
         return os.path.join(self.get_target_private_dir(target), src)
 
     def get_unity_source_file(self, target, suffix):
+        # There is a potential conflict here, but it is unlikely that
+        # anyone both enables unity builds and has a file called foo-unity.cpp.
         osrc = target.name + '-unity.' + suffix
         return mesonlib.File.from_built_file(self.get_target_private_dir(target), osrc)
 
@@ -224,8 +241,11 @@ class Backend:
                                os.path.join('dummyprefixdir', fromdir))
 
     def flatten_object_list(self, target, proj_dir_to_build_root=''):
+        return self._flatten_object_list(target, target.get_objects(), proj_dir_to_build_root)
+
+    def _flatten_object_list(self, target, objects, proj_dir_to_build_root):
         obj_list = []
-        for obj in target.get_objects():
+        for obj in objects:
             if isinstance(obj, str):
                 o = os.path.join(proj_dir_to_build_root,
                                  self.build_to_src, target.get_subdir(), obj)
@@ -233,7 +253,9 @@ class Backend:
             elif isinstance(obj, mesonlib.File):
                 obj_list.append(obj.rel_to_builddir(self.build_to_src))
             elif isinstance(obj, build.ExtractedObjects):
-                obj_list += self.determine_ext_objs(target, obj, proj_dir_to_build_root)
+                if obj.recursive:
+                    obj_list += self._flatten_object_list(obj.target, obj.objlist, proj_dir_to_build_root)
+                obj_list += self.determine_ext_objs(obj, proj_dir_to_build_root)
             else:
                 raise MesonException('Unknown data type in object list.')
         return obj_list
@@ -346,15 +368,11 @@ class Backend:
                 result += [rp]
         return result
 
-    def object_filename_from_source(self, target, source, is_unity):
+    def object_filename_from_source(self, target, source):
         assert isinstance(source, mesonlib.File)
         build_dir = self.environment.get_build_dir()
         rel_src = source.rel_to_builddir(self.build_to_src)
 
-        if (not self.environment.is_source(rel_src) or
-           self.environment.is_header(rel_src)) and not is_unity:
-            return None
-
         # foo.vala files compile down to foo.c and then foo.c.o, not foo.vala.o
         if rel_src.endswith(('.vala', '.gs')):
             # See description in generate_vala_compile for this logic.
@@ -364,8 +382,6 @@ class Backend:
                 rel_src = os.path.relpath(rel_src, self.get_target_private_dir(target))
             else:
                 rel_src = os.path.basename(rel_src)
-            if is_unity:
-                return 'meson-generated_' + rel_src[:-5] + '.c.' + self.environment.get_object_suffix()
             # A meson- prefixed directory is reserved; hopefully no-one creates a file name with such a weird prefix.
             source = 'meson-generated_' + rel_src[:-5] + '.c'
         elif source.is_built:
@@ -383,28 +399,42 @@ class Backend:
                                          os.path.join(self.environment.get_source_dir(), target.get_subdir()))
         return source.replace('/', '_').replace('\\', '_') + '.' + self.environment.get_object_suffix()
 
-    def determine_ext_objs(self, target, extobj, proj_dir_to_build_root):
+    def determine_ext_objs(self, extobj, proj_dir_to_build_root):
         result = []
+
+        # Merge sources and generated sources
+        sources = list(extobj.srclist)
+        for gensrc in extobj.genlist:
+            for s in gensrc.get_outputs():
+                path = self.get_target_generated_dir(extobj.target, gensrc, s)
+                dirpart, fnamepart = os.path.split(path)
+                sources.append(File(True, dirpart, fnamepart))
+
+        # Filter out headers and all non-source files
+        sources = [s for s in sources if self.environment.is_source(s) and not self.environment.is_header(s)]
+
+        # extobj could contain only objects and no sources
+        if not sources:
+            return result
+
         targetdir = self.get_target_private_dir(extobj.target)
+
         # With unity builds, there's just one object that contains all the
         # sources, and we only support extracting all the objects in this mode,
         # so just return that.
-        if self.is_unity(target):
-            comp = get_compiler_for_source(extobj.target.compilers.values(),
-                                           extobj.srclist[0])
-            # There is a potential conflict here, but it is unlikely that
-            # anyone both enables unity builds and has a file called foo-unity.cpp.
-            osrc = self.get_unity_source_file(extobj.target,
-                                              comp.get_default_suffix())
-            objname = self.object_filename_from_source(extobj.target, osrc, True)
-            objname = objname.replace('/', '_').replace('\\', '_')
+        if self.is_unity(extobj.target):
+            compsrcs = classify_unity_sources(extobj.target.compilers.values(), sources)
+            sources = []
+            for comp in compsrcs.keys():
+                osrc = self.get_unity_source_file(extobj.target,
+                                                  comp.get_default_suffix())
+                sources.append(osrc)
+
+        for osrc in sources:
+            objname = self.object_filename_from_source(extobj.target, osrc)
             objpath = os.path.join(proj_dir_to_build_root, targetdir, objname)
-            return [objpath]
-        for osrc in extobj.srclist:
-            objname = self.object_filename_from_source(extobj.target, osrc, False)
-            if objname:
-                objpath = os.path.join(proj_dir_to_build_root, targetdir, objname)
-                result.append(objpath)
+            result.append(objpath)
+
         return result
 
     def get_pch_include_args(self, compiler, target):
@@ -449,7 +479,7 @@ class Backend:
         # starting from hard-coded defaults followed by build options and so on.
         commands = CompilerArgs(compiler)
 
-        copt_proxy = OptionOverrideProxy(target.option_overrides, self.environment.coredata.compiler_options)
+        copt_proxy = self.get_compiler_options_for_target(target)
         # First, the trivial ones that are impossible to override.
         #
         # Add -nostdinc/-nostdinc++ if needed; can't be overridden
@@ -516,12 +546,13 @@ class Backend:
             # pkg-config puts the thread flags itself via `Cflags:`
             if dep.need_threads():
                 commands += compiler.thread_flags(self.environment)
+            elif dep.need_openmp():
+                commands += compiler.openmp_flags()
         # Fortran requires extra include directives.
         if compiler.language == 'fortran':
             for lt in target.link_targets:
-                priv_dir = os.path.join(self.get_target_dir(lt), lt.get_basename() + lt.type_suffix())
-                incflag = compiler.get_include_args(priv_dir, False)
-                commands += incflag
+                priv_dir = self.get_target_private_dir(lt)
+                commands += compiler.get_include_args(priv_dir, False)
         return commands
 
     def build_target_link_arguments(self, compiler, deps):
@@ -696,6 +727,9 @@ class Backend:
                 if not isinstance(arg, (build.CustomTarget, build.BuildTarget)):
                     continue
                 result[arg.get_id()] = arg
+            for dep in t.depends:
+                assert isinstance(dep, (build.CustomTarget, build.BuildTarget))
+                result[dep.get_id()] = dep
         return result
 
     def get_custom_target_provided_libraries(self, target):
