@@ -14,7 +14,8 @@
 
 import copy, os, re
 from collections import OrderedDict
-import itertools
+import itertools, pathlib
+import pickle
 
 from . import environment
 from . import dependencies
@@ -22,67 +23,67 @@ from . import mlog
 from .mesonlib import File, MesonException, listify, extract_as_list
 from .mesonlib import typeslistify, stringlistify, classify_unity_sources
 from .mesonlib import get_filenames_templates_dict, substitute_values
-from .mesonlib import for_windows, for_darwin, for_cygwin
+from .mesonlib import for_windows, for_darwin, for_cygwin, for_android, has_path_sep
 from .compilers import is_object, clike_langs, sort_clike, lang_suffixes
 
-known_basic_kwargs = {'install': True,
-                      'c_pch': True,
-                      'cpp_pch': True,
-                      'c_args': True,
-                      'objc_args': True,
-                      'objcpp_args': True,
-                      'cpp_args': True,
-                      'cs_args': True,
-                      'vala_args': True,
-                      'fortran_args': True,
-                      'd_args': True,
-                      'd_import_dirs': True,
-                      'd_unittest': True,
-                      'd_module_versions': True,
-                      'java_args': True,
-                      'rust_args': True,
-                      'link_args': True,
-                      'link_depends': True,
-                      'link_with': True,
-                      'link_whole': True,
-                      'implicit_include_directories': True,
-                      'include_directories': True,
-                      'dependencies': True,
-                      'install_dir': True,
-                      'main_class': True,
-                      'name_suffix': True,
-                      'gui_app': True,
-                      'extra_files': True,
-                      'install_rpath': True,
-                      'build_rpath': True,
-                      'resources': True,
-                      'sources': True,
-                      'objects': True,
-                      'native': True,
-                      'build_by_default': True,
-                      'override_options': True,
-                      }
+pch_kwargs = set(['c_pch', 'cpp_pch'])
 
-# These contain kwargs supported by both static and shared libraries. These are
-# combined here because a library() call might be shared_library() or
-# static_library() at runtime based on the configuration.
-# FIXME: Find a way to pass that info down here so we can have proper target
-# kwargs checking when specifically using shared_library() or static_library().
-known_lib_kwargs = known_basic_kwargs.copy()
-known_lib_kwargs.update({'version': True, # Only for shared libs
-                         'soversion': True, # Only for shared libs
-                         'name_prefix': True,
-                         'vs_module_defs': True, # Only for shared libs
-                         'vala_header': True,
-                         'vala_vapi': True,
-                         'vala_gir': True,
-                         'pic': True, # Only for static libs
-                         'rust_crate_type': True, # Only for Rust libs
-                         })
+lang_arg_kwargs = set([
+    'c_args',
+    'cpp_args',
+    'd_args',
+    'd_import_dirs',
+    'd_unittest',
+    'd_module_versions',
+    'fortran_args',
+    'java_args',
+    'objc_args',
+    'objcpp_args',
+    'rust_args',
+    'vala_args',
+    'cs_args',
+])
 
-known_exe_kwargs = known_basic_kwargs.copy()
-known_exe_kwargs.update({'implib': True,
-                         })
+vala_kwargs = set(['vala_header', 'vala_gir', 'vala_vapi'])
+rust_kwargs = set(['rust_crate_type'])
+cs_kwargs = set(['resources', 'cs_args'])
+
+buildtarget_kwargs = set([
+    'build_by_default',
+    'build_rpath',
+    'dependencies',
+    'extra_files',
+    'gui_app',
+    'link_with',
+    'link_whole',
+    'link_args',
+    'link_depends',
+    'implicit_include_directories',
+    'include_directories',
+    'install',
+    'install_rpath',
+    'install_dir',
+    'name_prefix',
+    'name_suffix',
+    'native',
+    'objects',
+    'override_options',
+    'sources',
+])
+
+known_build_target_kwargs = (
+    buildtarget_kwargs |
+    lang_arg_kwargs |
+    pch_kwargs |
+    vala_kwargs |
+    rust_kwargs |
+    cs_kwargs)
+
+known_exe_kwargs = known_build_target_kwargs | {'implib', 'export_dynamic'}
+known_shlib_kwargs = known_build_target_kwargs | {'version', 'soversion', 'vs_module_defs'}
+known_shmod_kwargs = known_build_target_kwargs
+known_stlib_kwargs = known_build_target_kwargs | {'pic'}
+known_jar_kwargs = known_exe_kwargs | {'main_class'}
 
 class InvalidArguments(MesonException):
     pass
@@ -112,6 +113,7 @@ class Build:
         self.static_linker = None
         self.static_cross_linker = None
         self.subprojects = {}
+        self.subproject_dir = ''
         self.install_scripts = []
         self.postconf_scripts = []
         self.install_dirs = []
@@ -119,6 +121,8 @@ class Build:
         self.dep_manifest = {}
         self.cross_stdlibs = {}
         self.test_setups = {}
+        self.find_overrides = {}
+        self.searched_programs = set() # The list of all programs that have been searched for.
 
     def add_compiler(self, compiler):
         if self.static_linker is None and compiler.needs_static_linker():
@@ -137,6 +141,9 @@ class Build:
     def get_project(self):
         return self.projects['']
 
+    def get_subproject_dir(self):
+        return self.subproject_dir
+
     def get_targets(self):
         return self.targets
 
@@ -206,39 +213,49 @@ class ExtractedObjects:
     '''
     Holds a list of sources for which the objects must be extracted
     '''
-    def __init__(self, target, srclist, is_unity):
+    def __init__(self, target, srclist=[], genlist=[], objlist=[], recursive=True):
         self.target = target
+        self.recursive = recursive
         self.srclist = srclist
-        if is_unity:
+        self.genlist = genlist
+        self.objlist = objlist
+        if self.target.is_unity:
             self.check_unity_compatible()
 
     def __repr__(self):
         r = '<{0} {1!r}: {2}>'
         return r.format(self.__class__.__name__, self.target.name, self.srclist)
 
+    def classify_all_sources(self, sources, generated_sources):
+        # Merge sources and generated sources
+        sources = list(sources)
+        for gensrc in generated_sources:
+            for s in gensrc.get_outputs():
+                # We cannot know the path where this source will be generated,
+                # but all we need here is the file extension to determine the
+                # compiler.
+                sources.append(s)
+
+        # Filter out headers and all non-source files
+        sources = [s for s in sources if environment.is_source(s) and not environment.is_header(s)]
+
+        return classify_unity_sources(self.target.compilers.values(), sources)
+
     def check_unity_compatible(self):
         # Figure out if the extracted object list is compatible with a Unity
         # build. When we're doing a Unified build, we go through the sources,
         # and create a single source file from each subset of the sources that
         # can be compiled with a specific compiler. Then we create one object
-        # from each unified source file.
-        # If the list of sources for which we want objects is the same as the
-        # list of sources that go into each unified build, we're good.
-        srclist_set = set(self.srclist)
-        # Objects for all the sources are required, so we're compatible
-        if srclist_set == set(self.target.sources):
-            return
-        # Check if the srclist is a subset (of the target's sources) that is
-        # going to form a unified source file and a single object
-        compsrcs = classify_unity_sources(self.target.compilers.values(),
-                                          self.target.sources)
-        for srcs in compsrcs.values():
-            if srclist_set == set(srcs):
-                return
-        msg = 'Single object files can not be extracted in Unity builds. ' \
-              'You can only extract all the object files at once.'
-        raise MesonException(msg)
+        # from each unified source file. So for each compiler we can either
+        # extra all its sources or none.
+        cmpsrcs = self.classify_all_sources(self.target.sources, self.target.generated)
+        extracted_cmpsrcs = self.classify_all_sources(self.srclist, self.genlist)
 
+        for comp, srcs in extracted_cmpsrcs.items():
+            if set(srcs) != set(cmpsrcs[comp]):
+                raise MesonException('Single object files can not be extracted '
+                                     'in Unity builds. You can only extract all '
+                                     'the object files for each compiler at once.')
 
 class EnvironmentVariables:
     def __init__(self):
@@ -280,7 +297,7 @@ class EnvironmentVariables:
 
 class Target:
     def __init__(self, name, subdir, subproject, build_by_default):
-        if '/' in name or '\\' in name:
+        if has_path_sep(name):
             # Fix failing test 53 when this becomes an error.
             mlog.warning('''Target "%s" has a path separator in its name.
 This is not supported, it can cause unexpected failures and will become
@@ -302,11 +319,16 @@ a hard error in the future.''' % name)
     def get_id(self):
         # This ID must also be a valid file name on all OSs.
         # It should also avoid shell metacharacters for obvious
-        # reasons.
-        base = self.name + self.type_suffix()
-        if self.subproject == '':
-            return base
-        return self.subproject + '@@' + base
+        # reasons. '@' is not used as often as '_' in source code names.
+        # In case of collisions consider using checksums.
+        # FIXME replace with assert when slash in names is prohibited
+        name_part = self.name.replace('/', '@').replace('\\', '@')
+        assert not has_path_sep(self.type_suffix())
+        myid = name_part + self.type_suffix()
+        if self.subdir:
+            subdir_part = self.subdir.replace('/', '@').replace('\\', '@')
+            myid = subdir_part + '@@' + myid
+        return myid
 
     def process_kwargs(self, kwargs):
         if 'build_by_default' in kwargs:
@@ -329,6 +351,8 @@ a hard error in the future.''' % name)
 
 
 class BuildTarget(Target):
+    known_kwargs = known_build_target_kwargs
+
     def __init__(self, name, subdir, subproject, is_cross, sources, objects, environment, kwargs):
         super().__init__(name, subdir, subproject, True)
         self.is_cross = is_cross
@@ -354,6 +378,7 @@ class BuildTarget(Target):
         self.extra_args = {}
         self.generated = []
         self.extra_files = []
+        self.d_features = {}
         # Sources can be:
         # 1. Pre-existing source files in the source tree
         # 2. Pre-existing sources generated by configure_file in the build tree
@@ -363,14 +388,15 @@ class BuildTarget(Target):
         # 1. Pre-existing objects provided by the user with the `objects:` kwarg
         # 2. Compiled objects created by and extracted from another target
         self.process_objectlist(objects)
-        self.process_compilers()
         self.process_kwargs(kwargs, environment)
         self.check_unknown_kwargs(kwargs)
+        self.process_compilers()
         if not any([self.sources, self.generated, self.objects, self.link_whole]):
             raise InvalidArguments('Build target %s has no sources.' % name)
         self.process_compilers_late()
         self.validate_sources()
         self.validate_cross_install(environment)
+        self.check_module_linking()
 
     def __lt__(self, other):
         return self.get_id() < other.get_id()
@@ -386,7 +412,7 @@ class BuildTarget(Target):
     def check_unknown_kwargs(self, kwargs):
         # Override this method in derived classes that have more
         # keywords.
-        self.check_unknown_kwargs_int(kwargs, known_basic_kwargs)
+        self.check_unknown_kwargs_int(kwargs, self.known_kwargs)
 
     def check_unknown_kwargs_int(self, kwargs, known_kwargs):
         unknowns = []
@@ -498,6 +524,13 @@ class BuildTarget(Target):
                 # which is what we need.
                 if not is_object(s):
                     sources.append(s)
+        for d in self.external_deps:
+            if hasattr(d, 'held_object'):
+                d = d.held_object
+            for s in d.sources:
+                if isinstance(s, (str, File)):
+                    sources.append(s)
+
         # Sources that were used to create our extracted objects
         for o in self.objects:
             if not isinstance(o, ExtractedObjects):
@@ -609,13 +642,15 @@ class BuildTarget(Target):
             if not isinstance(src, str):
                 raise MesonException('Object extraction arguments must be strings.')
             src = File(False, self.subdir, src)
+            # FIXME: It could be a generated source
             if src not in self.sources:
                 raise MesonException('Tried to extract unknown source %s.' % src)
             obj_src.append(src)
-        return ExtractedObjects(self, obj_src, self.is_unity)
+        return ExtractedObjects(self, obj_src)
 
-    def extract_all_objects(self):
-        return ExtractedObjects(self, self.sources, self.is_unity)
+    def extract_all_objects(self, recursive=True):
+        return ExtractedObjects(self, self.sources, self.generated, self.objects,
+                                recursive)
 
     def get_all_link_deps(self):
         return self.get_transitive_link_deps()
@@ -649,10 +684,6 @@ just like those detected with the dependency() function.''')
             self.link(linktarget)
         lwhole = extract_as_list(kwargs, 'link_whole')
         for linktarget in lwhole:
-            # Sorry for this hack. Keyword targets are kept in holders
-            # in kwargs. Unpack here without looking at the exact type.
-            if hasattr(linktarget, "held_object"):
-                linktarget = linktarget.held_object
             self.link_whole(linktarget)
 
         c_pchlist, cpp_pchlist, clist, cpplist, cslist, valalist,  objclist, objcpplist, fortranlist, rustlist \
@@ -681,12 +712,14 @@ just like those detected with the dependency() function.''')
         dfeature_versions = kwargs.get('d_module_versions', None)
         if dfeature_versions:
             dfeatures['versions'] = dfeature_versions
-        dfeature_import_dirs = kwargs.get('d_import_dirs', None)
-        if dfeature_import_dirs:
+        if 'd_import_dirs' in kwargs:
+            dfeature_import_dirs = extract_as_list(kwargs, 'd_import_dirs', unholder=True)
+            for d in dfeature_import_dirs:
+                if not isinstance(d, IncludeDirs):
+                    raise InvalidArguments('Arguments to d_import_dirs must be include_directories.')
             dfeatures['import_dirs'] = dfeature_import_dirs
         if dfeatures:
-            if 'd' in self.compilers:
-                self.add_compiler_args('d', self.compilers['d'].get_feature_args(dfeatures))
+            self.d_features = dfeatures
 
         self.link_args = extract_as_list(kwargs, 'link_args')
         for i in self.link_args:
@@ -784,12 +817,16 @@ This will become a hard error in a future Meson release.''')
     def get_extra_args(self, language):
         return self.extra_args.get(language, [])
 
-    def get_dependencies(self):
+    def get_dependencies(self, exclude=None):
         transitive_deps = []
-        for t in self.link_targets + self.link_whole_targets:
+        if exclude is None:
+            exclude = []
+        for t in itertools.chain(self.link_targets, self.link_whole_targets):
+            if t in transitive_deps or t in exclude:
+                continue
             transitive_deps.append(t)
             if isinstance(t, StaticLibrary):
-                transitive_deps += t.get_dependencies()
+                transitive_deps += t.get_dependencies(transitive_deps + exclude)
         return transitive_deps
 
     def get_source_subdir(self):
@@ -830,13 +867,16 @@ This will become a hard error in a future Meson release.''')
                 self.add_include_dirs(dep.include_directories)
                 for l in dep.libraries:
                     self.link(l)
-                # Those parts that are external.
-                extpart = dependencies.InternalDependency('undefined',
-                                                          [],
-                                                          dep.compile_args,
-                                                          dep.link_args,
-                                                          [], [], [])
-                self.external_deps.append(extpart)
+                for l in dep.whole_libraries:
+                    self.link_whole(l)
+                if dep.compile_args or dep.link_args:
+                    # Those parts that are external.
+                    extpart = dependencies.InternalDependency('undefined',
+                                                              [],
+                                                              dep.compile_args,
+                                                              dep.link_args,
+                                                              [], [], [], [])
+                    self.external_deps.append(extpart)
                 # Deps of deps.
                 self.add_deps(dep.ext_deps)
             elif isinstance(dep, dependencies.Dependency):
@@ -1011,6 +1051,15 @@ You probably should put it in link_with instead.''')
     def is_linkable_target(self):
         return False
 
+    def check_module_linking(self):
+        '''
+        Warn if shared modules are linked with target: (link_with) #2865
+        '''
+        for link_target in self.link_targets:
+            if isinstance(link_target, SharedModule):
+                mlog.warning('''target links against shared modules. This is not
+recommended as it can lead to undefined behaviour on some platforms''')
+                return
 
 class Generator:
     def __init__(self, args, kwargs):
@@ -1053,7 +1102,7 @@ class Generator:
                 raise InvalidArguments('"output" may only contain strings.')
             if '@BASENAME@' not in rule and '@PLAINNAME@' not in rule:
                 raise InvalidArguments('Every element of "output" must contain @BASENAME@ or @PLAINNAME@.')
-            if '/' in rule or '\\' in rule:
+            if has_path_sep(rule):
                 raise InvalidArguments('"outputs" must not contain a directory separator.')
         if len(outputs) > 1:
             for o in outputs:
@@ -1064,7 +1113,7 @@ class Generator:
             depfile = kwargs['depfile']
             if not isinstance(depfile, str):
                 raise InvalidArguments('Depfile must be a string.')
-            if os.path.split(depfile)[1] != depfile:
+            if os.path.basename(depfile) != depfile:
                 raise InvalidArguments('Depfile must be a plain filename without a subdirectory.')
             self.depfile = depfile
         if 'capture' in kwargs:
@@ -1074,48 +1123,71 @@ class Generator:
             self.capture = capture
 
     def get_base_outnames(self, inname):
-        plainname = os.path.split(inname)[1]
+        plainname = os.path.basename(inname)
         basename = os.path.splitext(plainname)[0]
-        return [x.replace('@BASENAME@', basename).replace('@PLAINNAME@', plainname) for x in self.outputs]
+        bases = [x.replace('@BASENAME@', basename).replace('@PLAINNAME@', plainname) for x in self.outputs]
+        return bases
 
     def get_dep_outname(self, inname):
         if self.depfile is None:
             raise InvalidArguments('Tried to get dep name for rule that does not have dependency file defined.')
-        plainname = os.path.split(inname)[1]
+        plainname = os.path.basename(inname)
         basename = os.path.splitext(plainname)[0]
         return self.depfile.replace('@BASENAME@', basename).replace('@PLAINNAME@', plainname)
 
     def get_arglist(self, inname):
-        plainname = os.path.split(inname)[1]
+        plainname = os.path.basename(inname)
         basename = os.path.splitext(plainname)[0]
         return [x.replace('@BASENAME@', basename).replace('@PLAINNAME@', plainname) for x in self.arglist]
 
-    def process_files(self, name, files, state, extra_args=[]):
-        output = GeneratedList(self, extra_args=extra_args)
+    def is_parent_path(self, parent, trial):
+        relpath = pathlib.PurePath(trial).relative_to(parent)
+        return relpath.parts[0] != '..' # For subdirs we can only go "down".
+
+    def process_files(self, name, files, state, preserve_path_from=None, extra_args=[]):
+        output = GeneratedList(self, state.subdir, preserve_path_from, extra_args=extra_args)
         for f in files:
             if isinstance(f, str):
                 f = File.from_source_file(state.environment.source_dir, state.subdir, f)
             elif not isinstance(f, File):
                 raise InvalidArguments('{} arguments must be strings or files not {!r}.'.format(name, f))
-            output.add_file(f)
+            if preserve_path_from:
+                abs_f = f.absolute_path(state.environment.source_dir, state.environment.build_dir)
+                if not self.is_parent_path(preserve_path_from, abs_f):
+                    raise InvalidArguments('When using preserve_path_from, all input files must be in a subdirectory of the given dir.')
+            output.add_file(f, state)
         return output
 
 
 class GeneratedList:
-    def __init__(self, generator, extra_args=[]):
+    def __init__(self, generator, subdir, preserve_path_from=None, extra_args=[]):
         if hasattr(generator, 'held_object'):
             generator = generator.held_object
         self.generator = generator
         self.name = self.generator.exe
+        self.subdir = subdir
         self.infilelist = []
         self.outfilelist = []
         self.outmap = {}
         self.extra_depends = []
+        self.preserve_path_from = preserve_path_from
         self.extra_args = extra_args
 
-    def add_file(self, newfile):
+    def add_preserved_path_segment(self, infile, outfiles, state):
+        result = []
+        in_abs = infile.absolute_path(state.environment.source_dir, state.environment.build_dir)
+        assert(os.path.isabs(self.preserve_path_from))
+        rel = os.path.relpath(in_abs, self.preserve_path_from)
+        path_segment = os.path.dirname(rel)
+        for of in outfiles:
+            result.append(os.path.join(path_segment, of))
+        return result
+
+    def add_file(self, newfile, state):
         self.infilelist.append(newfile)
         outfiles = self.generator.get_base_outnames(newfile.fname)
+        if self.preserve_path_from:
+            outfiles = self.add_preserved_path_segment(newfile, outfiles, state)
         self.outfilelist += outfiles
         self.outmap[newfile] = outfiles
 
@@ -1135,6 +1207,8 @@ class GeneratedList:
         return self.extra_args
 
 class Executable(BuildTarget):
+    known_kwargs = known_exe_kwargs
+
     def __init__(self, name, subdir, subproject, is_cross, sources, objects, environment, kwargs):
         super().__init__(name, subdir, subproject, is_cross, sources, objects, environment, kwargs)
         # Unless overridden, executables have no suffix or prefix. Except on
@@ -1147,7 +1221,11 @@ class Executable(BuildTarget):
                     for_cygwin(is_cross, environment) or 'cs' in self.compilers):
                 self.suffix = 'exe'
             else:
-                self.suffix = ''
+                if ('c' in self.compilers and self.compilers['c'].get_id().startswith('arm') or
+                   'cpp' in self.compilers and self.compilers['cpp'].get_id().startswith('arm')):
+                    self.suffix = 'axf'
+                else:
+                    self.suffix = ''
         self.filename = self.name
         if self.suffix:
             self.filename += '.' + self.suffix
@@ -1160,29 +1238,36 @@ class Executable(BuildTarget):
         # The import library that GCC would generate (and prefer)
         self.gcc_import_filename = None
 
-        # if implib appears, this target is linkwith:-able, but that only means
-        # something on Windows platforms.
-        self.is_linkwithable = False
-        if 'implib' in kwargs and kwargs['implib']:
+        # Check for export_dynamic
+        self.export_dynamic = False
+        if kwargs.get('export_dynamic'):
+            if not isinstance(kwargs['export_dynamic'], bool):
+                raise InvalidArguments('"export_dynamic" keyword argument must be a boolean')
+            self.export_dynamic = True
+        if kwargs.get('implib'):
+            self.export_dynamic = True
+        if self.export_dynamic and kwargs.get('implib') is False:
+            raise InvalidArguments('"implib" keyword argument must not be false for if "export_dynamic" is true')
+
+        # If using export_dynamic, set the import library name
+        if self.export_dynamic:
             implib_basename = self.name + '.exe'
-            if not isinstance(kwargs['implib'], bool):
+            if not isinstance(kwargs.get('implib', False), bool):
                 implib_basename = kwargs['implib']
-            self.is_linkwithable = True
             if for_windows(is_cross, environment) or for_cygwin(is_cross, environment):
                 self.vs_import_filename = '{0}.lib'.format(implib_basename)
                 self.gcc_import_filename = 'lib{0}.a'.format(implib_basename)
-
                 if self.get_using_msvc():
                     self.import_filename = self.vs_import_filename
                 else:
                     self.import_filename = self.gcc_import_filename
 
+        # Only linkwithable if using export_dynamic
+        self.is_linkwithable = self.export_dynamic
+
     def type_suffix(self):
         return "@exe"
 
-    def check_unknown_kwargs(self, kwargs):
-        self.check_unknown_kwargs_int(kwargs, known_exe_kwargs)
-
     def get_import_filename(self):
         """
         The name of the import library that will be outputted by the compiler
@@ -1200,6 +1285,8 @@ class Executable(BuildTarget):
         return self.is_linkwithable
 
 class StaticLibrary(BuildTarget):
+    known_kwargs = known_stlib_kwargs
+
     def __init__(self, name, subdir, subproject, is_cross, sources, objects, environment, kwargs):
         if 'pic' not in kwargs and 'b_staticpic' in environment.coredata.base_options:
             kwargs['pic'] = environment.coredata.base_options['b_staticpic'].value
@@ -1238,9 +1325,6 @@ class StaticLibrary(BuildTarget):
     def type_suffix(self):
         return "@sta"
 
-    def check_unknown_kwargs(self, kwargs):
-        self.check_unknown_kwargs_int(kwargs, known_lib_kwargs)
-
     def process_kwargs(self, kwargs, environment):
         super().process_kwargs(kwargs, environment)
         if 'rust_crate_type' in kwargs:
@@ -1254,6 +1338,8 @@ class StaticLibrary(BuildTarget):
         return True
 
 class SharedLibrary(BuildTarget):
+    known_kwargs = known_shlib_kwargs
+
     def __init__(self, name, subdir, subproject, is_cross, sources, objects, environment, kwargs):
         self.soversion = None
         self.ltversion = None
@@ -1359,6 +1445,11 @@ class SharedLibrary(BuildTarget):
             else:
                 # libfoo.dylib
                 self.filename_tpl = '{0.prefix}{0.name}.{0.suffix}'
+        elif for_android(is_cross, env):
+            prefix = 'lib'
+            suffix = 'so'
+            # Android doesn't support shared_library versioning
+            self.filename_tpl = '{0.prefix}{0.name}.{0.suffix}'
         else:
             prefix = 'lib'
             suffix = 'so'
@@ -1380,25 +1471,32 @@ class SharedLibrary(BuildTarget):
 
     def process_kwargs(self, kwargs, environment):
         super().process_kwargs(kwargs, environment)
-        # Shared library version
-        if 'version' in kwargs:
-            self.ltversion = kwargs['version']
-            if not isinstance(self.ltversion, str):
-                raise InvalidArguments('Shared library version needs to be a string, not ' + type(self.ltversion).__name__)
-            if not re.fullmatch(r'[0-9]+(\.[0-9]+){0,2}', self.ltversion):
-                raise InvalidArguments('Invalid Shared library version "{0}". Must be of the form X.Y.Z where all three are numbers. Y and Z are optional.'.format(self.ltversion))
-        # Try to extract/deduce the soversion
-        if 'soversion' in kwargs:
-            self.soversion = kwargs['soversion']
-            if isinstance(self.soversion, int):
-                self.soversion = str(self.soversion)
-            if not isinstance(self.soversion, str):
-                raise InvalidArguments('Shared library soversion is not a string or integer.')
-        elif self.ltversion:
-            # library version is defined, get the soversion from that
-            # We replicate what Autotools does here and take the first
-            # number of the version by default.
-            self.soversion = self.ltversion.split('.')[0]
+
+        if not for_android(self.is_cross, self.environment):
+            supports_versioning = True
+        else:
+            supports_versioning = False
+
+        if supports_versioning:
+            # Shared library version
+            if 'version' in kwargs:
+                self.ltversion = kwargs['version']
+                if not isinstance(self.ltversion, str):
+                    raise InvalidArguments('Shared library version needs to be a string, not ' + type(self.ltversion).__name__)
+                if not re.fullmatch(r'[0-9]+(\.[0-9]+){0,2}', self.ltversion):
+                    raise InvalidArguments('Invalid Shared library version "{0}". Must be of the form X.Y.Z where all three are numbers. Y and Z are optional.'.format(self.ltversion))
+            # Try to extract/deduce the soversion
+            if 'soversion' in kwargs:
+                self.soversion = kwargs['soversion']
+                if isinstance(self.soversion, int):
+                    self.soversion = str(self.soversion)
+                if not isinstance(self.soversion, str):
+                    raise InvalidArguments('Shared library soversion is not a string or integer.')
+            elif self.ltversion:
+                # library version is defined, get the soversion from that
+                # We replicate what Autotools does here and take the first
+                # number of the version by default.
+                self.soversion = self.ltversion.split('.')[0]
         # Visual Studio module-definitions file
         if 'vs_module_defs' in kwargs:
             path = kwargs['vs_module_defs']
@@ -1430,9 +1528,6 @@ class SharedLibrary(BuildTarget):
             else:
                 raise InvalidArguments('Invalid rust_crate_type "{0}": must be a string.'.format(rust_crate_type))
 
-    def check_unknown_kwargs(self, kwargs):
-        self.check_unknown_kwargs_int(kwargs, known_lib_kwargs)
-
     def get_import_filename(self):
         """
         The name of the import library that will be outputted by the compiler
@@ -1488,6 +1583,8 @@ class SharedLibrary(BuildTarget):
 # A shared library that is meant to be used with dlopen rather than linking
 # into something else.
 class SharedModule(SharedLibrary):
+    known_kwargs = known_shmod_kwargs
+
     def __init__(self, name, subdir, subproject, is_cross, sources, objects, environment, kwargs):
         if 'version' in kwargs:
             raise MesonException('Shared modules must not specify the version kwarg.')
@@ -1497,19 +1594,20 @@ class SharedModule(SharedLibrary):
         self.import_filename = None
 
 class CustomTarget(Target):
-    known_kwargs = {'input': True,
-                    'output': True,
-                    'command': True,
-                    'capture': False,
-                    'install': True,
-                    'install_dir': True,
-                    'build_always': True,
-                    'depends': True,
-                    'depend_files': True,
-                    'depfile': True,
-                    'build_by_default': True,
-                    'override_options': True,
-                    }
+    known_kwargs = set([
+        'input',
+        'output',
+        'command',
+        'capture',
+        'install',
+        'install_dir',
+        'build_always',
+        'depends',
+        'depend_files',
+        'depfile',
+        'build_by_default',
+        'override_options',
+    ])
 
     def __init__(self, name, subdir, subproject, kwargs, absolute_paths=False):
         super().__init__(name, subdir, subproject, False)
@@ -1603,7 +1701,11 @@ class CustomTarget(Target):
         for i in self.outputs:
             if not(isinstance(i, str)):
                 raise InvalidArguments('Output argument not a string.')
-            if '/' in i:
+            if i == '':
+                raise InvalidArguments('Output must not be empty.')
+            if i.strip() == '':
+                raise InvalidArguments('Output must not consist only of whitespace.')
+            if has_path_sep(i):
                 raise InvalidArguments('Output must not contain a path segment.')
             if '@INPUT@' in i or '@INPUT0@' in i:
                 m = 'Output cannot contain @INPUT@ or @INPUT0@, did you ' \
@@ -1625,7 +1727,7 @@ class CustomTarget(Target):
             depfile = kwargs['depfile']
             if not isinstance(depfile, str):
                 raise InvalidArguments('Depfile must be a string.')
-            if os.path.split(depfile)[1] != depfile:
+            if os.path.basename(depfile) != depfile:
                 raise InvalidArguments('Depfile must be a plain filename without a subdirectory.')
             self.depfile = depfile
         self.command = self.flatten_command(kwargs['command'])
@@ -1739,6 +1841,8 @@ class RunTarget(Target):
         return "@run"
 
 class Jar(BuildTarget):
+    known_kwargs = known_jar_kwargs
+
     def __init__(self, name, subdir, subproject, is_cross, sources, objects, environment, kwargs):
         super().__init__(name, subdir, subproject, is_cross, sources, objects, environment, kwargs)
         for s in self.sources:
@@ -1761,7 +1865,6 @@ class Jar(BuildTarget):
         # All jar targets are installable.
         pass
 
-
 class CustomTargetIndex:
 
     """A special opaque object returned by indexing a CustomTarget. This object
@@ -1831,13 +1934,19 @@ class ConfigurationData:
 # A bit poorly named, but this represents plain data files to copy
 # during install.
 class Data:
-    def __init__(self, sources, install_dir, install_mode=None):
+    def __init__(self, sources, install_dir, install_mode=None, rename=None):
         self.sources = sources
         self.install_dir = install_dir
         self.install_mode = install_mode
         self.sources = listify(self.sources)
         for s in self.sources:
             assert(isinstance(s, File))
+        if rename is None:
+            self.rename = [os.path.basename(f.fname) for f in self.sources]
+        else:
+            self.rename = stringlistify(rename)
+            if len(self.rename) != len(self.sources):
+                raise MesonException('Size of rename argument is different from number of sources')
 
 class RunScript(dict):
     def __init__(self, script, args):
@@ -1872,3 +1981,22 @@ def get_sources_string_names(sources):
         else:
             raise AssertionError('Unknown source type: {!r}'.format(s))
     return names
+
+def load(build_dir):
+    filename = os.path.join(build_dir, 'meson-private', 'build.dat')
+    load_fail_msg = 'Build data file {!r} is corrupted. Try with a fresh build tree.'.format(filename)
+    nonexisting_fail_msg = 'No such build data file as "{!r}".'.format(filename)
+    try:
+        with open(filename, 'rb') as f:
+            obj = pickle.load(f)
+    except FileNotFoundError:
+        raise MesonException(nonexisting_fail_msg)
+    except pickle.UnpicklingError:
+        raise MesonException(load_fail_msg)
+    if not isinstance(obj, Build):
+        raise MesonException(load_fail_msg)
+    return obj
+
+def save(obj, filename):
+    with open(filename, 'wb') as f:
+        pickle.dump(obj, f)
