@@ -42,9 +42,9 @@ d_feature_args = {'gcc':  {'unittest': '-funittest',
                   }
 
 class DCompiler(Compiler):
-    def __init__(self, exelist, version, is_cross):
+    def __init__(self, exelist, version, is_cross, **kwargs):
         self.language = 'd'
-        super().__init__(exelist, version)
+        super().__init__(exelist, version, **kwargs)
         self.id = 'unknown'
         self.is_cross = is_cross
 
@@ -89,11 +89,11 @@ class DCompiler(Compiler):
     def get_std_shared_lib_link_args(self):
         return ['-shared']
 
-    def get_soname_args(self, prefix, shlib_name, suffix, path, soversion, is_shared_module):
+    def get_soname_args(self, prefix, shlib_name, suffix, path, soversion, version, is_shared_module):
         # FIXME: Make this work for Windows, MacOS and cross-compiling
-        return get_gcc_soname_args(GCC_STANDARD, prefix, shlib_name, suffix, path, soversion, is_shared_module)
+        return get_gcc_soname_args(GCC_STANDARD, prefix, shlib_name, suffix, path, soversion, version, is_shared_module)
 
-    def get_feature_args(self, kwargs):
+    def get_feature_args(self, kwargs, build_to_src):
         res = []
         if 'unittest' in kwargs:
             unittest = kwargs.pop('unittest')
@@ -122,8 +122,16 @@ class DCompiler(Compiler):
             import_dir_arg = d_feature_args[self.id]['import_dir']
             if not import_dir_arg:
                 raise EnvironmentException('D compiler %s does not support the "string import directories" feature.' % self.name_string())
-            for d in import_dirs:
-                res.append('{0}{1}'.format(import_dir_arg, d))
+            for idir_obj in import_dirs:
+                basedir = idir_obj.get_curdir()
+                for idir in idir_obj.get_incdirs():
+                    # Avoid superfluous '/.' at the end of paths when d is '.'
+                    if idir not in ('', '.'):
+                        expdir = os.path.join(basedir, idir)
+                    else:
+                        expdir = basedir
+                    srctreedir = os.path.join(build_to_src, expdir)
+                    res.append('{0}{1}'.format(import_dir_arg, srctreedir))
 
         if kwargs:
             raise EnvironmentException('Unknown D compiler feature(s) selected: %s' % ', '.join(kwargs.keys()))
@@ -224,8 +232,8 @@ class DCompiler(Compiler):
 
 
 class GnuDCompiler(DCompiler):
-    def __init__(self, exelist, version, is_cross):
-        DCompiler.__init__(self, exelist, version, is_cross)
+    def __init__(self, exelist, version, is_cross, **kwargs):
+        DCompiler.__init__(self, exelist, version, is_cross, **kwargs)
         self.id = 'gcc'
         default_warn_args = ['-Wall', '-Wdeprecated']
         self.warn_args = {'1': default_warn_args,
@@ -274,8 +282,8 @@ class GnuDCompiler(DCompiler):
 
 
 class LLVMDCompiler(DCompiler):
-    def __init__(self, exelist, version, is_cross):
-        DCompiler.__init__(self, exelist, version, is_cross)
+    def __init__(self, exelist, version, is_cross, **kwargs):
+        DCompiler.__init__(self, exelist, version, is_cross, **kwargs)
         self.id = 'llvm'
         self.base_options = ['b_coverage', 'b_colorout']
 
@@ -328,8 +336,8 @@ class LLVMDCompiler(DCompiler):
 
 
 class DmdDCompiler(DCompiler):
-    def __init__(self, exelist, version, is_cross):
-        DCompiler.__init__(self, exelist, version, is_cross)
+    def __init__(self, exelist, version, is_cross, **kwargs):
+        DCompiler.__init__(self, exelist, version, is_cross, **kwargs)
         self.id = 'dmd'
         self.base_options = ['b_coverage', 'b_colorout']
 
