@@ -1,5 +1,4 @@
-
-# Copyright 2012-2017 The Meson development team
+# Copyright 2012-2018 The Meson development team
 
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -13,23 +12,32 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
+from . import mlog
 import pickle, os, uuid
 import sys
 from pathlib import PurePath
 from collections import OrderedDict
-from .mesonlib import MesonException, commonpath
+from .mesonlib import MesonException
 from .mesonlib import default_libdir, default_libexecdir, default_prefix
 import ast
+import argparse
 
-version = '0.44.1'
+version = '0.46.0'
 backendlist = ['ninja', 'vs', 'vs2010', 'vs2015', 'vs2017', 'xcode']
 
+default_yielding = False
+
 class UserOption:
-    def __init__(self, name, description, choices):
+    def __init__(self, name, description, choices, yielding):
         super().__init__()
         self.name = name
         self.choices = choices
         self.description = description
+        if yielding is None:
+            yielding = default_yielding
+        if not isinstance(yielding, bool):
+            raise MesonException('Value of "yielding" must be a boolean.')
+        self.yielding = yielding
 
     # Check that the input is a valid value and return the
     # "cleaned" or "native" version. For example the Boolean
@@ -37,76 +45,69 @@ class UserOption:
     def validate_value(self, value):
         raise RuntimeError('Derived option class did not override validate_value.')
 
+    def set_value(self, newvalue):
+        self.value = self.validate_value(newvalue)
+
 class UserStringOption(UserOption):
-    def __init__(self, name, description, value, choices=None):
-        super().__init__(name, description, choices)
+    def __init__(self, name, description, value, choices=None, yielding=None):
+        super().__init__(name, description, choices, yielding)
         self.set_value(value)
 
-    def validate(self, value):
+    def validate_value(self, value):
         if not isinstance(value, str):
             raise MesonException('Value "%s" for string option "%s" is not a string.' % (str(value), self.name))
-
-    def set_value(self, newvalue):
-        self.validate(newvalue)
-        self.value = newvalue
-
-    def validate_value(self, value):
-        self.validate(value)
         return value
 
 class UserBooleanOption(UserOption):
-    def __init__(self, name, description, value):
-        super().__init__(name, description, [True, False])
+    def __init__(self, name, description, value, yielding=None):
+        super().__init__(name, description, [True, False], yielding)
         self.set_value(value)
 
-    def tobool(self, thing):
-        if isinstance(thing, bool):
-            return thing
-        if thing.lower() == 'true':
-            return True
-        if thing.lower() == 'false':
-            return False
-        raise MesonException('Value %s is not boolean (true or false).' % thing)
-
-    def set_value(self, newvalue):
-        self.value = self.tobool(newvalue)
-
     def __bool__(self):
         return self.value
 
     def validate_value(self, value):
-        return self.tobool(value)
+        if isinstance(value, bool):
+            return value
+        if value.lower() == 'true':
+            return True
+        if value.lower() == 'false':
+            return False
+        raise MesonException('Value %s is not boolean (true or false).' % value)
 
 class UserIntegerOption(UserOption):
-    def __init__(self, name, description, min_value, max_value, value):
-        super().__init__(name, description, [True, False])
+    def __init__(self, name, description, min_value, max_value, value, yielding=None):
+        super().__init__(name, description, [True, False], yielding)
         self.min_value = min_value
         self.max_value = max_value
         self.set_value(value)
+        c = []
+        if min_value is not None:
+            c.append('>=' + str(min_value))
+        if max_value is not None:
+            c.append('<=' + str(max_value))
+        self.choices = ', '.join(c)
 
-    def set_value(self, newvalue):
-        if isinstance(newvalue, str):
-            newvalue = self.toint(newvalue)
-        if not isinstance(newvalue, int):
+    def validate_value(self, value):
+        if isinstance(value, str):
+            value = self.toint(value)
+        if not isinstance(value, int):
             raise MesonException('New value for integer option is not an integer.')
-        if self.min_value is not None and newvalue < self.min_value:
-            raise MesonException('New value %d is less than minimum value %d.' % (newvalue, self.min_value))
-        if self.max_value is not None and newvalue > self.max_value:
-            raise MesonException('New value %d is more than maximum value %d.' % (newvalue, self.max_value))
-        self.value = newvalue
+        if self.min_value is not None and value < self.min_value:
+            raise MesonException('New value %d is less than minimum value %d.' % (value, self.min_value))
+        if self.max_value is not None and value > self.max_value:
+            raise MesonException('New value %d is more than maximum value %d.' % (value, self.max_value))
+        return value
 
     def toint(self, valuestring):
         try:
             return int(valuestring)
-        except:
+        except ValueError:
             raise MesonException('Value string "%s" is not convertable to an integer.' % valuestring)
 
-    def validate_value(self, value):
-        return self.toint(value)
-
 class UserComboOption(UserOption):
-    def __init__(self, name, description, choices, value):
-        super().__init__(name, description, choices)
+    def __init__(self, name, description, choices, value, yielding=None):
+        super().__init__(name, description, choices, yielding)
         if not isinstance(self.choices, list):
             raise MesonException('Combo choices must be an array.')
         for i in self.choices:
@@ -114,23 +115,18 @@ class UserComboOption(UserOption):
                 raise MesonException('Combo choice elements must be strings.')
         self.set_value(value)
 
-    def set_value(self, newvalue):
-        if newvalue not in self.choices:
-            optionsstring = ', '.join(['"%s"' % (item,) for item in self.choices])
-            raise MesonException('Value "%s" for combo option "%s" is not one of the choices. Possible choices are: %s.' % (newvalue, self.name, optionsstring))
-        self.value = newvalue
-
     def validate_value(self, value):
         if value not in self.choices:
-            raise MesonException('Value %s not one of accepted values.' % value)
+            optionsstring = ', '.join(['"%s"' % (item,) for item in self.choices])
+            raise MesonException('Value "%s" for combo option "%s" is not one of the choices. Possible choices are: %s.' % (value, self.name, optionsstring))
         return value
 
 class UserArrayOption(UserOption):
     def __init__(self, name, description, value, **kwargs):
-        super().__init__(name, description, kwargs.get('choices', []))
-        self.set_value(value, user_input=False)
+        super().__init__(name, description, kwargs.get('choices', []), yielding=kwargs.get('yielding', None))
+        self.value = self.validate_value(value, user_input=False)
 
-    def validate(self, value, user_input):
+    def validate_value(self, value, user_input=True):
         # User input is for options defined on the command line (via -D
         # options). Users can put their input in as a comma separated
         # string, but for defining options in meson_options.txt the format
@@ -148,6 +144,9 @@ class UserArrayOption(UserOption):
                 newvalue = ast.literal_eval(value)
             else:
                 newvalue = [v.strip() for v in value.split(',')]
+                if len(set(newvalue)) != len(newvalue):
+                    mlog.log(mlog.red('DEPRECATION:'), '''Duplicated values in an array type is deprecated.
+This will become a hard error in the future.''')
         if not isinstance(newvalue, list):
             raise MesonException('"{0}" should be a string array, but it is not'.format(str(newvalue)))
         for i in newvalue:
@@ -160,13 +159,6 @@ class UserArrayOption(UserOption):
                     ', '.join(bad), ', '.join(self.choices)))
         return newvalue
 
-    def set_value(self, newvalue, user_input=True):
-        self.value = self.validate(newvalue, user_input)
-
-    def validate_value(self, value):
-        self.validate(value)
-        return value
-
 # This class contains all data that must persist over multiple
 # invocations of Meson. It is roughly the same thing as
 # cmakecache.
@@ -194,7 +186,6 @@ class CoreData:
         self.compilers = OrderedDict()
         self.cross_compilers = OrderedDict()
         self.deps = OrderedDict()
-        self.modules = {}
         # Only to print a warning if it changes between Meson invocations.
         self.pkgconf_envvar = os.environ.get('PKG_CONFIG_PATH', '')
 
@@ -206,17 +197,17 @@ class CoreData:
         (after resolving variables and ~), return that absolute path. Next,
         check if the file is relative to the current source dir. If the path
         still isn't resolved do the following:
-            Linux + BSD:
+            Windows:
+                - Error
+            *:
                 - $XDG_DATA_HOME/meson/cross (or ~/.local/share/meson/cross if
                   undefined)
                 - $XDG_DATA_DIRS/meson/cross (or
                   /usr/local/share/meson/cross:/usr/share/meson/cross if undefined)
                 - Error
-            *:
-                - Error
-        BSD follows the Linux path and will honor XDG_* if set. This simplifies
-        the implementation somewhat, especially since most BSD users wont set
-        those environment variables.
+
+        Non-Windows follows the Linux path and will honor XDG_* if set. This
+        simplifies the implementation somewhat.
         """
         if filename is None:
             return None
@@ -226,7 +217,7 @@ class CoreData:
         path_to_try = os.path.abspath(filename)
         if os.path.exists(path_to_try):
             return path_to_try
-        if sys.platform == 'linux' or 'bsd' in sys.platform.lower():
+        if sys.platform != 'win32':
             paths = [
                 os.environ.get('XDG_DATA_HOME', os.path.expanduser('~/.local/share')),
             ] + os.environ.get('XDG_DATA_DIRS', '/usr/local/share:/usr/share').split(':')
@@ -270,7 +261,7 @@ class CoreData:
             # commonpath will always return a path in the native format, so we
             # must use pathlib.PurePath to do the same conversion before
             # comparing.
-            if commonpath([value, prefix]) != str(PurePath(prefix)):
+            if os.path.commonpath([value, prefix]) != str(PurePath(prefix)):
                 m = 'The value of the {!r} option is {!r} which must be a ' \
                     'subdir of the prefix {!r}.\nNote that if you pass a ' \
                     'relative path, it is assumed to be a subdir of prefix.'
@@ -324,7 +315,8 @@ class CoreData:
                 return opt.validate_value(override_value)
         raise MesonException('Tried to validate unknown option %s.' % option_name)
 
-def load(filename):
+def load(build_dir):
+    filename = os.path.join(build_dir, 'meson-private', 'coredata.dat')
     load_fail_msg = 'Coredata file {!r} is corrupted. Try with a fresh build tree.'.format(filename)
     try:
         with open(filename, 'rb') as f:
@@ -338,7 +330,8 @@ def load(filename):
                              (obj.version, version))
     return obj
 
-def save(obj, filename):
+def save(obj, build_dir):
+    filename = os.path.join(build_dir, 'meson-private', 'coredata.dat')
     if obj.version != version:
         raise MesonException('Fatal version mismatch corruption.')
     with open(filename, 'wb') as f:
@@ -367,6 +360,20 @@ def get_builtin_option_description(optname):
     else:
         raise RuntimeError('Tried to get the description for an unknown builtin option \'%s\'.' % optname)
 
+def get_builtin_option_action(optname):
+    default = builtin_options[optname][2]
+    if default is True:
+        return 'store_false'
+    elif default is False:
+        return 'store_true'
+    return None
+
+def get_builtin_option_destination(optname):
+    optname = optname.replace('-', '_')
+    if optname == 'warnlevel':
+        return 'warning_level'
+    return optname
+
 def get_builtin_option_default(optname, prefix='', noneIfSuppress=False):
     if is_builtin_option(optname):
         o = builtin_options[optname]
@@ -384,6 +391,29 @@ def get_builtin_option_default(optname, prefix='', noneIfSuppress=False):
     else:
         raise RuntimeError('Tried to get the default value for an unknown builtin option \'%s\'.' % optname)
 
+def add_builtin_argument(p, name):
+    kwargs = {}
+    k = get_builtin_option_destination(name)
+    c = get_builtin_option_choices(k)
+    b = get_builtin_option_action(k)
+    h = get_builtin_option_description(k)
+    if not b:
+        h = h.rstrip('.') + ' (default: %s).' % get_builtin_option_default(k)
+    else:
+        kwargs['action'] = b
+    if c and not b:
+        kwargs['choices'] = c
+    default = get_builtin_option_default(k, noneIfSuppress=True)
+    if default is not None:
+        kwargs['default'] = default
+    else:
+        kwargs['default'] = argparse.SUPPRESS
+    p.add_argument('--' + name.replace('_', '-'), help=h, **kwargs)
+
+def register_builtin_arguments(parser):
+    for n in builtin_options:
+        add_builtin_argument(parser, n)
+
 builtin_options = {
     'buildtype':  [UserComboOption, 'Build type to use.', ['plain', 'debug', 'debugoptimized', 'release', 'minsize'], 'debug'],
     'strip':      [UserBooleanOption, 'Strip targets on install.', False],
@@ -404,7 +434,7 @@ builtin_options = {
     'werror':          [UserBooleanOption, 'Treat warnings as errors.', False],
     'warning_level':   [UserComboOption, 'Compiler warning level to use.', ['1', '2', '3'], '1'],
     'layout':          [UserComboOption, 'Build directory layout.', ['mirror', 'flat'], 'mirror'],
-    'default_library': [UserComboOption, 'Default library type.', ['shared', 'static'], 'shared'],
+    'default_library': [UserComboOption, 'Default library type.', ['shared', 'static', 'both'], 'shared'],
     'backend':         [UserComboOption, 'Backend to use.', backendlist, 'ninja'],
     'stdsplit':        [UserBooleanOption, 'Split stdout and stderr in test logs.', True],
     'errorlogs':       [UserBooleanOption, "Whether to print the logs from failing tests.", True],
