@@ -1,4 +1,4 @@
-# Copyright 2013-2017 The Meson development team
+# Copyright 2013-2018 The Meson development team
 
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -15,9 +15,9 @@
 # This file contains the detection logic for external dependencies.
 # Custom logic for several other packages are in separate files.
 
+import copy
 import os
 import re
-import sys
 import stat
 import shlex
 import shutil
@@ -62,21 +62,14 @@ class DependencyMethods(Enum):
 
 
 class Dependency:
-    def __init__(self, type_name, kwargs):
-        self.name = "null"
-        self.version = 'none'
-        self.language = None # None means C-like
-        self.is_found = False
-        self.type_name = type_name
-        self.compile_args = []
-        self.link_args = []
-        self.sources = []
+    @classmethod
+    def _process_method_kw(cls, kwargs):
         method = kwargs.get('method', 'auto')
         if method not in [e.value for e in DependencyMethods]:
             raise DependencyException('method {!r} is invalid'.format(method))
         method = DependencyMethods(method)
 
-        # This sets per-too config methods which are deprecated to to the new
+        # This sets per-tool config methods which are deprecated to to the new
         # generic CONFIG_TOOL value.
         if method in [DependencyMethods.SDLCONFIG, DependencyMethods.CUPSCONFIG,
                       DependencyMethods.PCAPCONFIG, DependencyMethods.LIBWMFCONFIG]:
@@ -89,14 +82,27 @@ class Dependency:
         # Set the detection method. If the method is set to auto, use any available method.
         # If method is set to a specific string, allow only that detection method.
         if method == DependencyMethods.AUTO:
-            self.methods = self.get_methods()
-        elif method in self.get_methods():
-            self.methods = [method]
+            methods = cls.get_methods()
+        elif method in cls.get_methods():
+            methods = [method]
         else:
             raise DependencyException(
                 'Unsupported detection method: {}, allowed methods are {}'.format(
                     method.value,
-                    mlog.format_list([x.value for x in [DependencyMethods.AUTO] + self.get_methods()])))
+                    mlog.format_list([x.value for x in [DependencyMethods.AUTO] + cls.get_methods()])))
+
+        return methods
+
+    def __init__(self, type_name, kwargs):
+        self.name = "null"
+        self.version = 'none'
+        self.language = None # None means C-like
+        self.is_found = False
+        self.type_name = type_name
+        self.compile_args = []
+        self.link_args = []
+        self.sources = []
+        self.methods = self._process_method_kw(kwargs)
 
     def __repr__(self):
         s = '<{0} {1}: {2}>'
@@ -116,7 +122,8 @@ class Dependency:
         As an example, gtest-all.cc when using GTest."""
         return self.sources
 
-    def get_methods(self):
+    @staticmethod
+    def get_methods():
         return [DependencyMethods.AUTO]
 
     def get_name(self):
@@ -128,6 +135,9 @@ class Dependency:
     def get_exe_args(self, compiler):
         return []
 
+    def need_openmp(self):
+        return False
+
     def need_threads(self):
         return False
 
@@ -137,9 +147,26 @@ class Dependency:
     def get_configtool_variable(self, variable_name):
         raise DependencyException('{!r} is not a config-tool dependency'.format(self.name))
 
+    def get_partial_dependency(self, *, compile_args=False, link_args=False,
+                               links=False, includes=False, sources=False):
+        """Create a new dependency that contains part of the parent dependency.
+
+        The following options can be inherited:
+            links -- all link_with arguemnts
+            includes -- all include_directory and -I/-isystem calls
+            sources -- any source, header, or generated sources
+            compile_args -- any compile args
+            link_args -- any link args
+
+        Additionally the new dependency will have the version parameter of it's
+        parent (if any) and the requested values of any dependencies will be
+        added as well.
+        """
+        RuntimeError('Unreachable code in partial_dependency called')
+
 
 class InternalDependency(Dependency):
-    def __init__(self, version, incdirs, compile_args, link_args, libraries, sources, ext_deps):
+    def __init__(self, version, incdirs, compile_args, link_args, libraries, whole_libraries, sources, ext_deps):
         super().__init__('internal', {})
         self.version = version
         self.is_found = True
@@ -147,6 +174,7 @@ class InternalDependency(Dependency):
         self.compile_args = compile_args
         self.link_args = link_args
         self.libraries = libraries
+        self.whole_libraries = whole_libraries
         self.sources = sources
         self.ext_deps = ext_deps
 
@@ -158,6 +186,21 @@ class InternalDependency(Dependency):
         raise DependencyException('Method "get_configtool_variable()" is '
                                   'invalid for an internal dependency')
 
+    def get_partial_dependency(self, *, compile_args=False, link_args=False,
+                               links=False, includes=False, sources=False):
+        compile_args = self.compile_args.copy() if compile_args else []
+        link_args = self.link_args.copy() if link_args else []
+        libraries = self.libraries.copy() if links else []
+        whole_libraries = self.whole_libraries.copy() if links else []
+        sources = self.sources.copy() if sources else []
+        includes = self.include_directories.copy() if includes else []
+        deps = [d.get_partial_dependency(
+            compile_args=compile_args, link_args=link_args, links=links,
+            includes=includes, sources=sources) for d in self.ext_deps]
+        return InternalDependency(
+            self.version, includes, compile_args, link_args, libraries,
+            whole_libraries, sources, deps)
+
 
 class ExternalDependency(Dependency):
     def __init__(self, type_name, environment, language, kwargs):
@@ -201,6 +244,26 @@ class ExternalDependency(Dependency):
     def get_compiler(self):
         return self.compiler
 
+    def get_partial_dependency(self, *, compile_args=False, link_args=False,
+                               links=False, includes=False, sources=False):
+        new = copy.copy(self)
+        if not compile_args:
+            new.compile_args = []
+        if not link_args:
+            new.link_args = []
+        if not sources:
+            new.sources = []
+
+        return new
+
+
+class NotFoundDependency(Dependency):
+    def __init__(self, environment):
+        super().__init__('not-found', {})
+        self.env = environment
+        self.name = 'not-found'
+        self.is_found = False
+
 
 class ConfigToolDependency(ExternalDependency):
 
@@ -213,6 +276,7 @@ class ConfigToolDependency(ExternalDependency):
     def __init__(self, name, environment, language, kwargs):
         super().__init__('config-tool', environment, language, kwargs)
         self.name = name
+        self.native = kwargs.get('native', False)
         self.tools = listify(kwargs.get('tools', self.tools))
 
         req_version = kwargs.get('version', None)
@@ -246,14 +310,17 @@ class ConfigToolDependency(ExternalDependency):
         # instantiated and returned. The reduce function (method) is also
         # attached, since python's pickle module won't be able to do anything
         # with this dynamically generated class otherwise.
-        def reduce(_):
-            return (cls.factory,
-                    (name, environment, language, kwargs, tools, tool_name))
+        def reduce(self):
+            return (cls._unpickle, (), self.__dict__)
         sub = type('{}Dependency'.format(name.capitalize()), (cls, ),
                    {'tools': tools, 'tool_name': tool_name, '__reduce__': reduce})
 
         return sub(name, environment, language, kwargs)
 
+    @classmethod
+    def _unpickle(cls):
+        return cls.__new__(cls)
+
     def find_config(self, versions=None):
         """Helper method that searchs for config tool binaries in PATH and
         returns the one that best matches the given version requirements.
@@ -261,8 +328,20 @@ class ConfigToolDependency(ExternalDependency):
         if not isinstance(versions, list) and versions is not None:
             versions = listify(versions)
 
+        if self.env.is_cross_build() and not self.native:
+            cross_file = self.env.cross_info.config['binaries']
+            try:
+                tools = [cross_file[self.tool_name]]
+            except KeyError:
+                mlog.warning('No entry for {0} specified in your cross file. '
+                             'Falling back to searching PATH. This may find a '
+                             'native version of {0}!'.format(self.tool_name))
+                tools = self.tools
+        else:
+            tools = self.tools
+
         best_match = (None, None)
-        for tool in self.tools:
+        for tool in tools:
             try:
                 p, out = Popen_safe([tool, '--version'])[:2]
             except (FileNotFoundError, PermissionError):
@@ -272,8 +351,8 @@ class ConfigToolDependency(ExternalDependency):
 
             out = self._sanitize_version(out.strip())
             # Some tools, like pcap-config don't supply a version, but also
-            # dont fail with --version, in that case just assume that there is
-            # only one verison and return it.
+            # don't fail with --version, in that case just assume that there is
+            # only one version and return it.
             if not out:
                 return (tool, 'none')
             if versions:
@@ -295,17 +374,18 @@ class ConfigToolDependency(ExternalDependency):
         """Helper method to print messages about the tool."""
         if self.config is None:
             if version is not None:
-                mlog.log('found {} {!r} but need:'.format(self.tool_name, version),
-                         req_version)
+                mlog.log('Found', mlog.bold(self.tool_name), repr(version),
+                         mlog.red('NO'), '(needed', req_version, ')')
             else:
-                mlog.log("No {} found; can't detect dependency".format(self.tool_name))
-            mlog.log('Dependency {} found:'.format(self.name), mlog.red('NO'))
+                mlog.log('Found', mlog.bold(self.tool_name), repr(req_version),
+                         mlog.red('NO'))
+            mlog.log('Dependency', mlog.bold(self.name), 'found:', mlog.red('NO'))
             if self.required:
                 raise DependencyException('Dependency {} not found'.format(self.name))
             return False
         mlog.log('Found {}:'.format(self.tool_name), mlog.bold(shutil.which(self.config)),
                  '({})'.format(version))
-        mlog.log('Dependency {} found:'.format(self.name), mlog.green('YES'))
+        mlog.log('Dependency', mlog.bold(self.name), 'found:', mlog.green('YES'))
         return True
 
     def get_config_value(self, args, stage):
@@ -318,7 +398,8 @@ class ConfigToolDependency(ExternalDependency):
             return []
         return shlex.split(out)
 
-    def get_methods(self):
+    @staticmethod
+    def get_methods():
         return [DependencyMethods.AUTO, DependencyMethods.CONFIG_TOOL]
 
     def get_configtool_variable(self, variable_name):
@@ -337,6 +418,8 @@ class PkgConfigDependency(ExternalDependency):
     # The class's copy of the pkg-config path. Avoids having to search for it
     # multiple times in the same Meson invocation.
     class_pkgbin = None
+    # We cache all pkg-config subprocess invocations to avoid redundant calls
+    pkgbin_cache = {}
 
     def __init__(self, name, environment, kwargs, language=None):
         super().__init__('pkgconfig', environment, language, kwargs)
@@ -356,9 +439,7 @@ class PkgConfigDependency(ExternalDependency):
                 pkgname = environment.cross_info.config['binaries']['pkgconfig']
                 potential_pkgbin = ExternalProgram(pkgname, silent=True)
                 if potential_pkgbin.found():
-                    # FIXME, we should store all pkg-configs in ExternalPrograms.
-                    # However that is too destabilizing a change to do just before release.
-                    self.pkgbin = potential_pkgbin.get_command()[0]
+                    self.pkgbin = potential_pkgbin
                     PkgConfigDependency.class_pkgbin = self.pkgbin
                 else:
                     mlog.debug('Cross pkg-config %s not found.' % potential_pkgbin.name)
@@ -380,7 +461,7 @@ class PkgConfigDependency(ExternalDependency):
             self.type_string = 'Native'
 
         mlog.debug('Determining dependency {!r} with pkg-config executable '
-                   '{!r}'.format(name, self.pkgbin))
+                   '{!r}'.format(name, self.pkgbin.get_path()))
         ret, self.version = self._call_pkgbin(['--modversion', name])
         if ret != 0:
             if self.required:
@@ -410,11 +491,22 @@ class PkgConfigDependency(ExternalDependency):
                     m = 'Invalid version of dependency, need {!r} {!r} found {!r}.'
                     raise DependencyException(m.format(name, not_found, self.version))
                 return
-        found_msg += [mlog.green('YES'), self.version]
-        # Fetch cargs to be used while using this dependency
-        self._set_cargs()
-        # Fetch the libraries and library paths needed for using this
-        self._set_libs()
+
+        try:
+            # Fetch cargs to be used while using this dependency
+            self._set_cargs()
+            # Fetch the libraries and library paths needed for using this
+            self._set_libs()
+            found_msg += [mlog.green('YES'), self.version]
+        except DependencyException as e:
+            if self.required:
+                raise
+            else:
+                self.compile_args = []
+                self.link_args = []
+                self.is_found = False
+                found_msg += [mlog.red('NO'), '; reason: {}'.format(str(e))]
+
         # Print the found message only at the very end because fetching cflags
         # and libs can also fail if other needed pkg-config files aren't found.
         if not self.silent:
@@ -425,12 +517,22 @@ class PkgConfigDependency(ExternalDependency):
         return s.format(self.__class__.__name__, self.name, self.is_found,
                         self.version_reqs)
 
-    def _call_pkgbin(self, args, env=None):
-        if not env:
-            env = os.environ
-        p, out = Popen_safe([self.pkgbin] + args, env=env)[0:2]
+    def _call_pkgbin_real(self, args, env):
+        p, out = Popen_safe(self.pkgbin.get_command() + args, env=env)[0:2]
         return p.returncode, out.strip()
 
+    def _call_pkgbin(self, args, env=None):
+        if env is None:
+            fenv = env
+            env = os.environ
+        else:
+            fenv = frozenset(env.items())
+        targs = tuple(args)
+        cache = PkgConfigDependency.pkgbin_cache
+        if (self.pkgbin, targs, fenv) not in cache:
+            cache[(self.pkgbin, targs, fenv)] = self._call_pkgbin_real(args, env)
+        return cache[(self.pkgbin, targs, fenv)]
+
     def _convert_mingw_paths(self, args):
         '''
         Both MSVC and native Python on Windows cannot handle MinGW-esque /c/foo
@@ -460,7 +562,13 @@ class PkgConfigDependency(ExternalDependency):
         return converted
 
     def _set_cargs(self):
-        ret, out = self._call_pkgbin(['--cflags', self.name])
+        env = None
+        if self.language == 'fortran':
+            # gfortran doesn't appear to look in system paths for INCLUDE files,
+            # so don't allow pkg-config to suppress -I flags for system paths
+            env = os.environ.copy()
+            env['PKG_CONFIG_ALLOW_SYSTEM_CFLAGS'] = '1'
+        ret, out = self._call_pkgbin(['--cflags', self.name], env=env)
         if ret != 0:
             raise DependencyException('Could not generate cargs for %s:\n\n%s' %
                                       (self.name, out))
@@ -543,10 +651,22 @@ class PkgConfigDependency(ExternalDependency):
                                           (self.type_string, self.name))
         else:
             variable = out.strip()
+
+            # pkg-config doesn't distinguish between empty and non-existent variables
+            # use the variable list to check for variable existence
+            if not variable:
+                ret, out = self._call_pkgbin(['--print-variables', self.name])
+                if not re.search(r'^' + variable_name + r'$', out, re.MULTILINE):
+                    if 'default' in kwargs:
+                        variable = kwargs['default']
+                    else:
+                        mlog.warning("pkgconfig variable '%s' not defined for dependency %s." % (variable_name, self.name))
+
         mlog.debug('Got pkgconfig variable %s : %s' % (variable_name, variable))
         return variable
 
-    def get_methods(self):
+    @staticmethod
+    def get_methods():
         return [DependencyMethods.PKGCONFIG]
 
     def check_pkgconfig(self):
@@ -555,21 +675,23 @@ class PkgConfigDependency(ExternalDependency):
             pkgbin = os.environ[evar].strip()
         else:
             pkgbin = 'pkg-config'
-        try:
-            p, out = Popen_safe([pkgbin, '--version'])[0:2]
-            if p.returncode != 0:
-                # Set to False instead of None to signify that we've already
-                # searched for it and not found it
+        pkgbin = ExternalProgram(pkgbin, silent=True)
+        if pkgbin.found():
+            try:
+                p, out = Popen_safe(pkgbin.get_command() + ['--version'])[0:2]
+                if p.returncode != 0:
+                    mlog.warning('Found pkg-config {!r} but couldn\'t run it'
+                                 ''.format(' '.join(pkgbin.get_command())))
+                    # Set to False instead of None to signify that we've already
+                    # searched for it and not found it
+                    pkgbin = False
+            except (FileNotFoundError, PermissionError):
                 pkgbin = False
-        except (FileNotFoundError, PermissionError):
+        else:
             pkgbin = False
-        if pkgbin and not os.path.isabs(pkgbin) and shutil.which(pkgbin):
-            # Sometimes shutil.which fails where Popen succeeds, so
-            # only find the abs path if it can be found by shutil.which
-            pkgbin = shutil.which(pkgbin)
         if not self.silent:
             if pkgbin:
-                mlog.log('Found pkg-config:', mlog.bold(pkgbin),
+                mlog.log('Found pkg-config:', mlog.bold(pkgbin.get_path()),
                          '(%s)' % out.strip())
             else:
                 mlog.log('Found Pkg-config:', mlog.red('NO'))
@@ -818,6 +940,15 @@ class ExternalLibrary(ExternalDependency):
             return []
         return self.link_args
 
+    def get_partial_dependency(self, *, compile_args=False, link_args=False,
+                               links=False, includes=False, sources=False):
+        # External library only has link_args, so ignore the rest of the
+        # interface.
+        new = copy.copy(self)
+        if not link_args:
+            new.link_args = []
+        return new
+
 
 class ExtraFrameworkDependency(ExternalDependency):
     def __init__(self, name, required, path, env, lang, kwargs):
@@ -895,7 +1026,12 @@ def find_external_dependency(name, env, kwargs):
     if lname in packages:
         if lname not in _packages_accept_language and 'language' in kwargs:
             raise DependencyException('%s dependency does not accept "language" keyword argument' % (lname, ))
-        dep = packages[lname](env, kwargs)
+        # Create the dependency object using a factory class method, if one
+        # exists, otherwise it is just constructed directly.
+        if getattr(packages[lname], '_factory', None):
+            dep = packages[lname]._factory(env, kwargs)
+        else:
+            dep = packages[lname](env, kwargs)
         if required and not dep.found():
             raise DependencyException('Dependency "%s" not found' % name)
         return dep
