@@ -12,12 +12,12 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
-import sys, stat, traceback, pickle, argparse
-import time, datetime
+import sys, stat, traceback, argparse
+import datetime
 import os.path
 from . import environment, interpreter, mesonlib
 from . import build
-from . import mconf, mintro, mtest, rewriter
+from . import mconf, mintro, mtest, rewriter, minit
 import platform
 from . import mlog, coredata
 from .mesonlib import MesonException
@@ -25,47 +25,9 @@ from .wrap import WrapMode, wraptool
 
 default_warning = '1'
 
-def add_builtin_argument(p, name, **kwargs):
-    k = kwargs.get('dest', name.replace('-', '_'))
-    c = coredata.get_builtin_option_choices(k)
-    b = True if kwargs.get('action', None) in ['store_true', 'store_false'] else False
-    h = coredata.get_builtin_option_description(k)
-    if not b:
-        h = h.rstrip('.') + ' (default: %s).' % coredata.get_builtin_option_default(k)
-    if c and not b:
-        kwargs['choices'] = c
-    default = coredata.get_builtin_option_default(k, noneIfSuppress=True)
-    if default is not None:
-        kwargs['default'] = default
-    else:
-        kwargs['default'] = argparse.SUPPRESS
-    p.add_argument('--' + name, help=h, **kwargs)
-
 def create_parser():
     p = argparse.ArgumentParser(prog='meson')
-    add_builtin_argument(p, 'prefix')
-    add_builtin_argument(p, 'libdir')
-    add_builtin_argument(p, 'libexecdir')
-    add_builtin_argument(p, 'bindir')
-    add_builtin_argument(p, 'sbindir')
-    add_builtin_argument(p, 'includedir')
-    add_builtin_argument(p, 'datadir')
-    add_builtin_argument(p, 'mandir')
-    add_builtin_argument(p, 'infodir')
-    add_builtin_argument(p, 'localedir')
-    add_builtin_argument(p, 'sysconfdir')
-    add_builtin_argument(p, 'localstatedir')
-    add_builtin_argument(p, 'sharedstatedir')
-    add_builtin_argument(p, 'backend')
-    add_builtin_argument(p, 'buildtype')
-    add_builtin_argument(p, 'strip', action='store_true')
-    add_builtin_argument(p, 'unity')
-    add_builtin_argument(p, 'werror', action='store_true')
-    add_builtin_argument(p, 'layout')
-    add_builtin_argument(p, 'default-library')
-    add_builtin_argument(p, 'warnlevel', dest='warning_level')
-    add_builtin_argument(p, 'stdsplit', action='store_false')
-    add_builtin_argument(p, 'errorlogs', action='store_false')
+    coredata.register_builtin_arguments(p)
     p.add_argument('--cross-file', default=None,
                    help='File describing cross compilation environment.')
     p.add_argument('-D', action='append', dest='projectoptions', default=[], metavar="option",
@@ -87,6 +49,25 @@ def wrapmodetype(string):
         msg = 'invalid argument {!r}, use one of {}'.format(string, msg)
         raise argparse.ArgumentTypeError(msg)
 
+def filter_builtin_options(args, original_args):
+    """Filter out any builtin arguments passed as -D options.
+
+    Error if an argument is passed with -- and -D
+    """
+    arguments = dict(p.split('=', 1) for p in args.projectoptions)
+    meson_opts = set(arguments).intersection(set(coredata.builtin_options))
+    if meson_opts:
+        for arg in meson_opts:
+            value = arguments[arg]
+            if any([a.startswith('--{}'.format(arg)) for a in original_args]):
+                raise MesonException(
+                    'Argument "{0}" passed as both --{0} and -D{0}, but only '
+                    'one is allowed'.format(arg))
+            setattr(args, coredata.get_builtin_option_destination(arg), value)
+
+            # Remove the builtin option from the project args values
+            args.projectoptions.remove('{}={}'.format(arg, value))
+
 class MesonApp:
 
     def __init__(self, dir1, dir2, script_launcher, handshake, options, original_cmd_line_args):
@@ -147,10 +128,8 @@ class MesonApp:
     def generate(self):
         env = environment.Environment(self.source_dir, self.build_dir, self.meson_script_launcher, self.options, self.original_cmd_line_args)
         mlog.initialize(env.get_log_dir())
-        try:
+        with mesonlib.BuildDirLock(self.build_dir):
             self._generate(env)
-        finally:
-            mlog.shutdown()
 
     def _generate(self, env):
         mlog.debug('Build started at', datetime.datetime.now().isoformat())
@@ -199,6 +178,7 @@ class MesonApp:
         mlog.log('Build machine cpu:', mlog.bold(intr.builtin['build_machine'].cpu_method([], {})))
         intr.run()
         try:
+            dumpfile = os.path.join(env.get_scratch_dir(), 'build.dat')
             # We would like to write coredata as late as possible since we use the existence of
             # this file to check if we generated the build file successfully. Since coredata
             # includes settings, the build files must depend on it and appear newer. However, due
@@ -207,16 +187,14 @@ class MesonApp:
             # possible, but before build files, and if any error occurs, delete it.
             cdf = env.dump_coredata()
             g.generate(intr)
-            dumpfile = os.path.join(env.get_scratch_dir(), 'build.dat')
-            with open(dumpfile, 'wb') as f:
-                pickle.dump(b, f)
+            build.save(b, dumpfile)
             # Post-conf scripts must be run after writing coredata or else introspection fails.
             g.run_postconf_scripts()
         except:
-            os.unlink(cdf)
+            if 'cdf' in locals():
+                os.unlink(cdf)
             raise
 
-
 def run_script_command(args):
     cmdname = args[0]
     cmdargs = args[1:]
@@ -279,8 +257,8 @@ def run_script_command(args):
     return cmdfunc(cmdargs)
 
 def run(original_args, mainfile=None):
-    if sys.version_info < (3, 4):
-        print('Meson works correctly only with python 3.4+.')
+    if sys.version_info < (3, 5):
+        print('Meson works correctly only with python 3.5+.')
         print('You have python %s.' % sys.version)
         print('Please update your environment')
         return 1
@@ -289,6 +267,13 @@ def run(original_args, mainfile=None):
         # First check if we want to run a subcommand.
         cmd_name = args[0]
         remaining_args = args[1:]
+        # "help" is a special case: Since printing of the help may be
+        # delegated to a subcommand, we edit cmd_name before executing
+        # the rest of the logic here.
+        if cmd_name == 'help':
+            remaining_args += ['--help']
+            args = remaining_args
+            cmd_name = args[0]
         if cmd_name == 'test':
             return mtest.run(remaining_args)
         elif cmd_name == 'setup':
@@ -296,18 +281,18 @@ def run(original_args, mainfile=None):
             # FALLTHROUGH like it's 1972.
         elif cmd_name == 'introspect':
             return mintro.run(remaining_args)
-        elif cmd_name == 'test':
-            return mtest.run(remaining_args)
         elif cmd_name == 'rewrite':
             return rewriter.run(remaining_args)
         elif cmd_name == 'configure':
             try:
                 return mconf.run(remaining_args)
             except MesonException as e:
-                mlog.log(mlog.red('\nError configuring project:'), e)
+                mlog.exception(e)
                 sys.exit(1)
         elif cmd_name == 'wrap':
             return wraptool.run(remaining_args)
+        elif cmd_name == 'init':
+            return minit.run(remaining_args)
         elif cmd_name == 'runpython':
             import runpy
             script_file = remaining_args[0]
@@ -322,8 +307,8 @@ def run(original_args, mainfile=None):
             try:
                 sys.exit(run_script_command(args[1:]))
             except MesonException as e:
-                mlog.log(mlog.red('\nError in {} helper script:'.format(script)))
-                mlog.log(e)
+                mlog.error('\nError in {} helper script:'.format(script))
+                mlog.exception(e)
                 sys.exit(1)
         args = args[2:]
         handshake = True
@@ -334,6 +319,7 @@ def run(original_args, mainfile=None):
 
     args = mesonlib.expand_arguments(args)
     options = parser.parse_args(args)
+    filter_builtin_options(options, args)
     args = options.directories
     if not args or len(args) > 2:
         # if there's a meson.build in the dir above, and not in the current
@@ -366,20 +352,20 @@ def run(original_args, mainfile=None):
         app.generate()
     except Exception as e:
         if isinstance(e, MesonException):
-            if hasattr(e, 'file') and hasattr(e, 'lineno') and hasattr(e, 'colno'):
-                mlog.log(mlog.red('\nMeson encountered an error in file %s, line %d, column %d:' % (e.file, e.lineno, e.colno)))
-            else:
-                mlog.log(mlog.red('\nMeson encountered an error:'))
-            # Error message
-            mlog.log(e)
+            mlog.exception(e)
             # Path to log file
+            mlog.shutdown()
             logfile = os.path.join(app.build_dir, environment.Environment.log_dir, mlog.log_fname)
             mlog.log("\nA full log can be found at", mlog.bold(logfile))
             if os.environ.get('MESON_FORCE_BACKTRACE'):
                 raise
+            return 1
         else:
             if os.environ.get('MESON_FORCE_BACKTRACE'):
                 raise
             traceback.print_exc()
-        return 1
+            return 2
+    finally:
+        mlog.shutdown()
+
     return 0
